@@ -356,7 +356,22 @@ static PyGetSetDef pygpgme_context_getsets[] = {
     { NULL, (getter)0, (setter)0 }
 };
 
-/* XXX: set_locale */
+static PyObject *
+pygpgme_context_set_engine_info(PyGpgmeContext *self, PyObject *args)
+{
+    int protocol;
+    const char *file_name, *home_dir;
+
+    if (!PyArg_ParseTuple(args, "izz", &protocol, &file_name, &home_dir))
+        return NULL;
+
+    if (pygpgme_check_error(gpgme_ctx_set_engine_info(self->ctx, protocol,
+                                                      file_name, home_dir)))
+        return NULL;
+
+    Py_RETURN_NONE;
+}
+
 static PyObject *
 pygpgme_context_set_locale(PyGpgmeContext *self, PyObject *args)
 {
@@ -372,11 +387,6 @@ pygpgme_context_set_locale(PyGpgmeContext *self, PyObject *args)
     Py_RETURN_NONE;
 }
 
-/* the following don't seem to be used */
-/* XXX: signers_clear */
-/* XXX: signers_add */
-/* XXX: signers_enum */
-
 static PyObject *
 pygpgme_context_get_key(PyGpgmeContext *self, PyObject *args)
 {
@@ -424,10 +434,17 @@ decode_encrypt_result(PyGpgmeContext *self)
 
     list = PyList_New(0);
     for (key = res->invalid_recipients; key != NULL; key = key->next) {
-        PyObject *item, *err;
+        PyObject *item, *py_fpr, *err;
 
+        if (key->fpr)
+            py_fpr = PyUnicode_DecodeASCII(key->fpr, strlen(key->fpr),
+                                           "replace");
+        else {
+            py_fpr = Py_None;
+            Py_INCREF(py_fpr);
+        }
         err = pygpgme_error_object(key->reason);
-        item = Py_BuildValue("(zN)", key->fpr, err);
+        item = Py_BuildValue("(NN)", py_fpr, err);
         PyList_Append(list, item);
         Py_DECREF(item);
     }
@@ -442,121 +459,107 @@ decode_encrypt_result(PyGpgmeContext *self)
 static PyObject *
 pygpgme_context_encrypt(PyGpgmeContext *self, PyObject *args)
 {
-    PyObject *py_recp, *py_plain, *py_cipher;
+    PyObject *py_recp, *py_plain, *py_cipher, *recp_seq = NULL, *result = NULL;
     int flags, i, length;
-    gpgme_key_t *recp;
-    gpgme_data_t plain, cipher;
+    gpgme_key_t *recp = NULL;
+    gpgme_data_t plain = NULL, cipher = NULL;
     gpgme_error_t err;
 
     if (!PyArg_ParseTuple(args, "OiOO", &py_recp, &flags,
                           &py_plain, &py_cipher))
-        return NULL;
+        goto end;
 
-    py_recp = PySequence_Fast(py_recp, "first argument must be a sequence");
-    if (py_recp == NULL)
-        return NULL;
+    if (py_recp != Py_None) {
+        recp_seq = PySequence_Fast(py_recp, "first argument must be a "
+                                   "sequence or None");
+        if (recp_seq == NULL)
+            goto end;
 
-    length = PySequence_Fast_GET_SIZE(py_recp);
-    recp = malloc((length + 1) * sizeof (gpgme_key_t));
-    for (i = 0; i < length; i++) {
-        PyObject *item = PySequence_Fast_GET_ITEM(py_recp, i);
+        length = PySequence_Fast_GET_SIZE(recp_seq);
+        recp = malloc((length + 1) * sizeof (gpgme_key_t));
+        for (i = 0; i < length; i++) {
+            PyObject *item = PySequence_Fast_GET_ITEM(recp_seq, i);
 
-        if (!PyObject_TypeCheck(item, &PyGpgmeKey_Type)) {
-            free(recp);
-            Py_DECREF(py_recp);
-            PyErr_SetString(PyExc_TypeError, "items in first argument must "
-                            "be gpgme.Key objects");
-            return NULL;
+            if (!PyObject_TypeCheck(item, &PyGpgmeKey_Type)) {
+                PyErr_SetString(PyExc_TypeError, "items in first argument "
+                                "must be gpgme.Key objects");
+                goto end;
+            }
+            recp[i] = ((PyGpgmeKey *)item)->key;
         }
-        recp[i] = ((PyGpgmeKey *)item)->key;
-    }
-    recp[i] = NULL;
-
-    if (pygpgme_data_new(&plain, py_plain)) {
-        free(recp);
-        Py_DECREF(py_recp);
-        return NULL;    
-    }
-    if (pygpgme_data_new(&cipher, py_cipher)) {
-        free(recp);
-        Py_DECREF(py_recp);
-        gpgme_data_release(plain);
-        return NULL;    
+        recp[i] = NULL;
     }
 
-    Py_BEGIN_ALLOW_THREADS;
-    err = gpgme_op_encrypt(self->ctx, recp, flags, plain, cipher);
+    if (pygpgme_data_new(&plain, py_plain))
+        goto end;
+    if (pygpgme_data_new(&cipher, py_cipher))
+        goto end;
+
+    Py_BEGIN_ALLOW_THREADS;    err = gpgme_op_encrypt(self->ctx, recp, flags, plain, cipher);
     Py_END_ALLOW_THREADS;
 
-    free(recp);
-    Py_DECREF(py_recp);
-    gpgme_data_release(plain);
-    gpgme_data_release(cipher);
-
     if (pygpgme_check_error(err)) {
         decode_encrypt_result(self);
-        return NULL;
+        goto end;
     }
 
-    Py_RETURN_NONE;
+    Py_INCREF(Py_None);
+    result = Py_None;
+
+ end:
+    if (recp != NULL)
+        free(recp);
+    Py_XDECREF(recp_seq);
+    if (plain != NULL)
+        gpgme_data_release(plain);
+    if (cipher != NULL)
+        gpgme_data_release(cipher);
+
+    return result;
 }
 
 static PyObject *
 pygpgme_context_encrypt_sign(PyGpgmeContext *self, PyObject *args)
 {
-    PyObject *py_recp, *py_plain, *py_cipher;
+    PyObject *py_recp, *py_plain, *py_cipher, *recp_seq = NULL, *result = NULL;
     int flags, i, length;
-    gpgme_key_t *recp;
-    gpgme_data_t plain, cipher;
+    gpgme_key_t *recp = NULL;
+    gpgme_data_t plain = NULL, cipher = NULL;
     gpgme_error_t err;
-    gpgme_sign_result_t result;
+    gpgme_sign_result_t sign_result;
 
     if (!PyArg_ParseTuple(args, "OiOO", &py_recp, &flags,
                           &py_plain, &py_cipher))
-        return NULL;
+        goto end;
 
-    py_recp = PySequence_Fast(py_recp, "first argument must be a sequence");
-    if (py_recp == NULL)
-        return NULL;
+    recp_seq = PySequence_Fast(py_recp, "first argument must be a sequence");
+    if (recp_seq == NULL)
+        goto end;
 
-    length = PySequence_Fast_GET_SIZE(py_recp);
+    length = PySequence_Fast_GET_SIZE(recp_seq);
     recp = malloc((length + 1) * sizeof (gpgme_key_t));
     for (i = 0; i < length; i++) {
-        PyObject *item = PySequence_Fast_GET_ITEM(py_recp, i);
+        PyObject *item = PySequence_Fast_GET_ITEM(recp_seq, i);
 
         if (!PyObject_TypeCheck(item, &PyGpgmeKey_Type)) {
-            free(recp);
-            Py_DECREF(py_recp);
-            PyErr_SetString(PyExc_TypeError, "items in first argument must "
-                            "be gpgme.Key objects");
-            return NULL;
+            PyErr_SetString(PyExc_TypeError, "items in first argument "
+                            "must be gpgme.Key objects");
+            goto end;
         }
         recp[i] = ((PyGpgmeKey *)item)->key;
     }
     recp[i] = NULL;
 
-    if (pygpgme_data_new(&plain, py_plain)) {
-        free(recp);
-        Py_DECREF(py_recp);
-        return NULL;    
-    }
-    if (pygpgme_data_new(&cipher, py_cipher)) {
-        free(recp);
-        Py_DECREF(py_recp);
-        gpgme_data_release(plain);
-        return NULL;    
-    }
+    if (pygpgme_data_new(&plain, py_plain))
+        goto end;
+    if (pygpgme_data_new(&cipher, py_cipher))
+        goto end;
 
     Py_BEGIN_ALLOW_THREADS;
     err = gpgme_op_encrypt_sign(self->ctx, recp, flags, plain, cipher);
     Py_END_ALLOW_THREADS;
 
-    free(recp);
-    Py_DECREF(py_recp);
-    gpgme_data_release(plain);
-    gpgme_data_release(cipher);
-
-    result = gpgme_op_sign_result(self->ctx);
+    sign_result = gpgme_op_sign_result(self->ctx);
 
     /* annotate exception */
     if (pygpgme_check_error(err)) {
@@ -569,36 +572,54 @@ pygpgme_context_encrypt_sign(PyGpgmeContext *self, PyObject *args)
         PyErr_Fetch(&err_type, &err_value, &err_traceback);
         PyErr_NormalizeException(&err_type, &err_value, &err_traceback);
 
-        if (result == NULL)
-            goto end;
+        if (sign_result == NULL)
+            goto error_end;
 
         if (!PyErr_GivenExceptionMatches(err_type, pygpgme_error))
-            goto end;
+            goto error_end;
 
         list = PyList_New(0);
-        for (key = result->invalid_signers; key != NULL; key = key->next) {
-            PyObject *item, *err;
+        for (key = sign_result->invalid_signers; key != NULL; key = key->next) {
+            PyObject *item, *py_fpr, *err;
 
+            if (key->fpr)
+                py_fpr = PyUnicode_DecodeASCII(key->fpr, strlen(key->fpr),
+                                               "replace");
+            else {
+                py_fpr = Py_None;
+                Py_INCREF(py_fpr);
+            }
             err = pygpgme_error_object(key->reason);
-            item = Py_BuildValue("(zN)", key->fpr, err);
+            item = Py_BuildValue("(NN)", py_fpr, err);
             PyList_Append(list, item);
             Py_DECREF(item);
         }
         PyObject_SetAttrString(err_value, "invalid_signers", list);
         Py_DECREF(list);
 
-        list = pygpgme_newsiglist_new(result->signatures);
+        list = pygpgme_newsiglist_new(sign_result->signatures);
         PyObject_SetAttrString(err_value, "signatures", list);
         Py_DECREF(list);
-    end:
+    error_end:
         PyErr_Restore(err_type, err_value, err_traceback);
-        return NULL;
+        goto end;
     }
 
-    if (result)
-        return pygpgme_newsiglist_new(result->signatures);
+    if (sign_result)
+        result = pygpgme_newsiglist_new(sign_result->signatures);
     else
-        return PyList_New(0);
+        result = PyList_New(0);
+
+ end:
+    if (recp != NULL)
+        free(recp);
+    Py_XDECREF(recp_seq);
+    if (plain != NULL)
+        gpgme_data_release(plain);
+    if (cipher != NULL)
+        gpgme_data_release(cipher);
+
+    return result;
 }
 
 static void
@@ -619,7 +640,9 @@ decode_decrypt_result(PyGpgmeContext *self)
         goto end;
 
     if (res->unsupported_algorithm) {
-        value = PyString_FromString(res->unsupported_algorithm);
+        value = PyUnicode_DecodeUTF8(res->unsupported_algorithm,
+                                     strlen(res->unsupported_algorithm),
+                                     "replace");
     } else {
         Py_INCREF(Py_None);
         value = Py_None;
@@ -655,7 +678,7 @@ pygpgme_context_decrypt(PyGpgmeContext *self, PyObject *args)
 
     if (pygpgme_data_new(&plain, py_plain)) {
         gpgme_data_release(cipher);
-        return NULL;    
+        return NULL;
     }
 
     Py_BEGIN_ALLOW_THREADS;
@@ -690,7 +713,7 @@ pygpgme_context_decrypt_verify(PyGpgmeContext *self, PyObject *args)
 
     if (pygpgme_data_new(&plain, py_plain)) {
         gpgme_data_release(cipher);
-        return NULL;    
+        return NULL;
     }
 
     Py_BEGIN_ALLOW_THREADS;
@@ -748,11 +771,11 @@ pygpgme_context_sign(PyGpgmeContext *self, PyObject *args)
         return NULL;
 
     if (pygpgme_data_new(&plain, py_plain))
-        return NULL;    
+        return NULL;
 
     if (pygpgme_data_new(&sig, py_sig)) {
         gpgme_data_release(plain);
-        return NULL;    
+        return NULL;
     }
 
     Py_BEGIN_ALLOW_THREADS;
@@ -781,10 +804,17 @@ pygpgme_context_sign(PyGpgmeContext *self, PyObject *args)
 
         list = PyList_New(0);
         for (key = result->invalid_signers; key != NULL; key = key->next) {
-            PyObject *item, *err;
+            PyObject *item, *py_fpr, *err;
 
+            if (key->fpr)
+                py_fpr = PyUnicode_DecodeASCII(key->fpr, strlen(key->fpr),
+                                               "replace");
+            else {
+                py_fpr = Py_None;
+                Py_INCREF(py_fpr);
+            }
             err = pygpgme_error_object(key->reason);
-            item = Py_BuildValue("(zN)", key->fpr, err);
+            item = Py_BuildValue("(NN)", py_fpr, err);
             PyList_Append(list, item);
             Py_DECREF(item);
         }
@@ -822,12 +852,12 @@ pygpgme_context_verify(PyGpgmeContext *self, PyObject *args)
     }
     if (pygpgme_data_new(&signed_text, py_signed_text)) {
         gpgme_data_release(sig);
-        return NULL;    
+        return NULL;
     }
     if (pygpgme_data_new(&plaintext, py_plaintext)) {
         gpgme_data_release(sig);
         gpgme_data_release(signed_text);
-        return NULL;    
+        return NULL;
     }
 
     Py_BEGIN_ALLOW_THREADS;
@@ -924,9 +954,9 @@ pygpgme_context_export(PyGpgmeContext *self, PyObject *args)
         Py_INCREF(py_pattern);
         pattern = NULL;
         patterns = NULL;
-    } else if (PyString_Check(py_pattern)) {
+    } else if (PyBytes_Check(py_pattern)) {
         Py_INCREF(py_pattern);
-        pattern = PyString_AsString(py_pattern);
+        pattern = PyBytes_AsString(py_pattern);
         patterns = NULL;
     } else {
         py_pattern = PySequence_Fast(py_pattern,
@@ -939,14 +969,14 @@ pygpgme_context_export(PyGpgmeContext *self, PyObject *args)
         for (i = 0; i < length; i++) {
             PyObject *item = PySequence_Fast_GET_ITEM(py_pattern, i);
 
-            if (!PyString_Check(item)) {
+            if (!PyBytes_Check(item)) {
                 PyErr_SetString(PyExc_TypeError,
                     "first argument must be a string or sequence of strings");
                 free(patterns);
                 Py_DECREF(py_pattern);
                 return NULL;
             }
-            patterns[i] = PyString_AsString(item);
+            patterns[i] = PyBytes_AsString(item);
         }
         patterns[i] = NULL;
     }
@@ -974,7 +1004,55 @@ pygpgme_context_export(PyGpgmeContext *self, PyObject *args)
     Py_RETURN_NONE;
 }
 
-// pygpgme_context_genkey
+static PyObject *
+pygpgme_context_genkey(PyGpgmeContext *self, PyObject *args)
+{
+    PyObject *py_pubkey = Py_None, *py_seckey = Py_None;
+    const char *parms;
+    gpgme_data_t pubkey = NULL, seckey = NULL;
+    PyObject *result;
+    gpgme_error_t err;
+
+    if (!PyArg_ParseTuple(args, "z|OO", &parms, &py_pubkey, &py_seckey))
+        return NULL;
+
+    if (pygpgme_data_new(&pubkey, py_pubkey))
+        return NULL;
+
+    if (pygpgme_data_new(&seckey, py_seckey)) {
+        gpgme_data_release(pubkey);
+        return NULL;
+    }
+
+    Py_BEGIN_ALLOW_THREADS;
+    err = gpgme_op_genkey(self->ctx, parms, pubkey, seckey);
+    Py_END_ALLOW_THREADS;
+
+    gpgme_data_release(seckey);
+    gpgme_data_release(pubkey);
+    result = pygpgme_genkey_result(self->ctx);
+
+    if (pygpgme_check_error(err)) {
+        PyObject *err_type, *err_value, *err_traceback;
+
+        PyErr_Fetch(&err_type, &err_value, &err_traceback);
+        PyErr_NormalizeException(&err_type, &err_value, &err_traceback);
+
+        if (!PyErr_GivenExceptionMatches(err_type, pygpgme_error))
+            goto end;
+
+        if (result != NULL) {
+            PyObject_SetAttrString(err_value, "result", result);
+            Py_DECREF(result);
+        }
+    end:
+        PyErr_Restore(err_type, err_value, err_traceback);
+        return NULL;
+    }
+
+    return (PyObject *) result;
+}
+
 
 static PyObject *
 pygpgme_context_delete(PyGpgmeContext *self, PyObject *args)
@@ -1083,9 +1161,9 @@ pygpgme_context_keylist(PyGpgmeContext *self, PyObject *args)
         Py_INCREF(py_pattern);
         pattern = NULL;
         patterns = NULL;
-    } else if (PyString_Check(py_pattern)) {
+    } else if (PyBytes_Check(py_pattern)) {
         Py_INCREF(py_pattern);
-        pattern = PyString_AsString(py_pattern);
+        pattern = PyBytes_AsString(py_pattern);
         patterns = NULL;
     } else {
         py_pattern = PySequence_Fast(py_pattern,
@@ -1098,14 +1176,14 @@ pygpgme_context_keylist(PyGpgmeContext *self, PyObject *args)
         for (i = 0; i < length; i++) {
             PyObject *item = PySequence_Fast_GET_ITEM(py_pattern, i);
 
-            if (!PyString_Check(item)) {
+            if (!PyBytes_Check(item)) {
                 PyErr_SetString(PyExc_TypeError,
                     "first argument must be a string or sequence of strings");
                 free(patterns);
                 Py_DECREF(py_pattern);
                 return NULL;
             }
-            patterns[i] = PyString_AsString(item);
+            patterns[i] = PyBytes_AsString(item);
         }
         patterns[i] = NULL;
     }
@@ -1136,6 +1214,7 @@ pygpgme_context_keylist(PyGpgmeContext *self, PyObject *args)
 // pygpgme_context_trustlist
 
 static PyMethodDef pygpgme_context_methods[] = {
+    { "set_engine_info", (PyCFunction)pygpgme_context_set_engine_info, METH_VARARGS },
     { "set_locale", (PyCFunction)pygpgme_context_set_locale, METH_VARARGS },
     { "get_key", (PyCFunction)pygpgme_context_get_key, METH_VARARGS },
     { "encrypt", (PyCFunction)pygpgme_context_encrypt, METH_VARARGS },
@@ -1146,7 +1225,7 @@ static PyMethodDef pygpgme_context_methods[] = {
     { "verify", (PyCFunction)pygpgme_context_verify, METH_VARARGS },
     { "import_", (PyCFunction)pygpgme_context_import, METH_VARARGS },
     { "export", (PyCFunction)pygpgme_context_export, METH_VARARGS },
-    // genkey
+    { "genkey", (PyCFunction)pygpgme_context_genkey, METH_VARARGS },
     { "delete", (PyCFunction)pygpgme_context_delete, METH_VARARGS },
     { "edit", (PyCFunction)pygpgme_context_edit, METH_VARARGS },
     { "card_edit", (PyCFunction)pygpgme_context_card_edit, METH_VARARGS },
@@ -1156,8 +1235,7 @@ static PyMethodDef pygpgme_context_methods[] = {
 };
 
 PyTypeObject PyGpgmeContext_Type = {
-    PyObject_HEAD_INIT(NULL)
-    0,
+    PyVarObject_HEAD_INIT(NULL, 0)
     "gpgme.Context",
     sizeof(PyGpgmeContext),
     .tp_flags = Py_TPFLAGS_DEFAULT,
